#!/usr/bin/lua
require "lunit"
log = require "syslog"
lfs = require "lfs"
json = require "cjson"
local socket = require "socket"
socket.unix = require "socket.unix"
posix = require "posix"

module("processor", package.seeall, lunit.testcase)

function setup()
	-- Setup before each test
	local currentdir = lfs.currentdir()
	package.path = package.path..";"..currentdir.."/share/?.lua;"..currentdir.."/etc/rules/?.lua;"..currentdir.."/etc/rules/custom/?.lua;"..currentdir.."/etc/?.lua;"
	require "config"

	-- Make stdout print to variable "nonstdout"
	-- Note!!! Does not work with print() (print can not be overriden). Specify output as io.stdout and :write to it.
	copy = {
		write = function(self, ...) nonstdout = ... end,
	}
	copy.__index = copy
	io.stdout = setmetatable({mode = "stdout"}, copy)
end

----------------------------------------------
-- Connects to and queries the socket
-- @param cmd string
-- @return string
----------------------------------------------
function connect_and_query_socket(cmd)
	cmd = cmd.."\n"
	local s, sock = pcall(socket.unix)
	if s then
		assert(sock.connect(sock, config.socket_path))
		assert(sock:send(cmd))
		local response = sock:receive()
		return response
	else
		return "die"
	end
end

----------------------------------------------
-- Maybe this test doesn't belong here BUT
-- at least it is not released to production.
----------------------------------------------
function test_no_debug_true_commited()
	assert_false(config.debug, "Do not commit config.debug = true!!!")
end

----------------------------------------------
-- Test for startup errors
-- This also tests that livestatus socket is initiated and config is set.
----------------------------------------------
function test_for_startup_errors()
	-- Hi-jack args
	arg[1] = nil
	local s, processor =  pcall(loadfile, "bin/synergy")
	local output = processor()
	assert_equal("Nothing to do", output, "Expected processor to start up with no errors and return \"Nothing to do\", got "..tostring(output))
end

----------------------------------------------
-- Test that nothing crashes when spool is empty
----------------------------------------------
function test_empty_rule_spool()
	-- Hi-jack args
	arg[1] = "--process_spool"
	-- Set new spool dir
	config.spool_dir = "test/var/"
	assert_equal(posix.stat(config.spool_dir, "type"), "directory", "Don't forget to create the spool dir prior to testing: "..config.spool_dir)
	local s, processor, m =  pcall(loadfile, "bin/synergy")
	assert(s, "Expected no errors with processor, got errors")
	local status = processor()
	assert_true(status, "Expected function to return true, no such luck I'm afraid")
end

----------------------------------------------
-- Test executing a named rule that doesn't exist
-- Rewrite to utilize socket
----------------------------------------------
function test_execute_named_rule_that_doesnt_exist()
	config.tree_dir = "etc/bps/"
	arg[1] = "nonexistingrule"
	local s, processor = pcall(loadfile, "bin/synergy")
	assert_true(s, "Named execution failed. Got: "..tostring(processor))
	local outcome = processor()
	assert_false(outcome, "There shouldn't have been a rule with supplied name")
	assert(nonstdout, "Function should have returned an error message but didn't")
	local s, decoded_result = pcall(json.decode, nonstdout)
	assert_false(s, "Decoding of json succeded which shouldn't be possible")
	assert_string(decoded_result, "There should be an error string here")
end

----------------------------------------------
-- Test executing a named rule that exists
-- Rewrite to utilize socket
----------------------------------------------
function test_execute_named_rule_that_does_exist()
	config.tree_dir = "etc/bps/"
	arg[1] = "Test-Drive"
	local s, processor = pcall(loadfile, "bin/synergy")
	assert_true(s, "Named execution failed. Got: "..tostring(processor))
	local outcome, full, top = processor()
	assert_true(outcome, "There should have been a rule with supplied name")
	local s, decoded_result = pcall(json.decode, full)
	assert_true(s, "Decoding of json failed")
	assert_table(decoded_result, "No proper json string was returned")
	local s, decoded_result = pcall(json.decode, top)
	assert_true(s, "Decoding of json failed")
	assert_table(decoded_result, "No proper json string was returned")
end

function test_save_report_data_new_state_is_ok()
	local old_service_data = {
		has_been_checked = 1,
		state = 2
	}
	local result = {
		status = 0
	}
	local object = {
		name = "my bsm node",
		materialized = {
			host = "my bsm host"
		}
	}
	local res = nil
	local timestamp = os.time()
	local out = {}
	function write_cb(data)
		-- nonsense, small name for easyness-to-refer-to in assertions
		-- later on
		out.a = data
		return true
	end
	assert_true(
		save_report_data(
			old_service_data,
			result,
			object,
			res,
			timestamp,
			write_cb
		)
	)
	assert_equal(timestamp, out.a.meta.timestamp)
	assert_equal("my bsm host", out.a.meta.host)
	assert_equal("my bsm node", out.a.meta.name)
	assert_equal(0, out.a.meta.state)
	assert_equal("", out.a.data)
end

function test_save_report_data_new_state_is_critical()
	local old_service_data = {
		has_been_checked = 1,
		state = 0
	}
	local result = {
		status = 2
	}
	local object = {
		name = "my bsm node",
		materialized = {
			host = "my bsm host"
		}
	}
	local res = {
		items = {
			hello = {
				result = {
					status = "CRITICAL"
				}
			}
		}
	}
	local timestamp = os.time()
	local out = {}
	function write_cb(data)
		-- nonsense, small name for easyness-to-refer-to in assertions
		-- later on
		out.a = data
		return true
	end
	assert_true(
		save_report_data(
			old_service_data,
			result,
			object,
			res,
			timestamp,
			write_cb
		)
	)
	assert_equal(timestamp, out.a.meta.timestamp)
	assert_equal("my bsm host", out.a.meta.host)
	assert_equal("my bsm node", out.a.meta.name)
	assert_equal(2, out.a.meta.state)
	assert_equal("CRITICAL", out.a.data["items"]["hello"]["status"], json.encode(out.a.data))
end

function test_save_report_data_first_check_result()
	local old_service_data = {
		has_been_checked = 0,
		state = 0
	}
	local result = {
		status = 0
	}
	local object = {
		name = "my bsm node",
		materialized = {
			host = "my bsm host"
		}
	}
	local res = {
		items = {
			hello = {
				result = {
					status = "OK"
				}
			}
		}
	}
	local timestamp = os.time()
	local out = {}
	function write_cb(parsed_data)
		-- nonsense, small name for easyness-to-refer-to in assertions
		-- later on
		out.a = parsed_data
		return true
	end
	assert_true(
		save_report_data(
			old_service_data,
			result,
			object,
			res,
			timestamp,
			write_cb
		)
	)
	assert_equal(timestamp, out.a.meta.timestamp)
	assert_equal("my bsm host", out.a.meta.host)
	assert_equal("my bsm node", out.a.meta.name)
	assert_equal(0, out.a.meta.state)
	assert_equal("", out.a.data, json.encode(out.a.data))
end

function test_save_report_data_no_state_change()
	local old_service_data = {
		has_been_checked = 1,
		state = 0
	}
	local result = {
		status = 0
	}
	local object = {
		name = "my bsm node",
		materialized = {
			host = "my bsm host"
		}
	}
	local res = {
		items = {
			hello = {
				result = {
					status = "OK"
				}
			}
		}
	}
	local timestamp = os.time()
	local out = {}
	function write_cb(parsed_data)
		-- nonsense, small name for easyness-to-refer-to in assertions
		-- later on
		out.a = parsed_data
		return true
	end
	assert_false(
		save_report_data(
			old_service_data,
			result,
			object,
			res,
			timestamp,
			write_cb
		)
	)
end

function test_save_report_data_without_data_from_livestatus_aka_materialized_object_not_found()
	local old_service_data = {} -- this represents an empty result from Livestatus
	local result = {
		status = 0
	}
	local object = {
		name = "my bsm node",
		materialized = {
			host = "my bsm host"
		}
	}
	local res = {
		items = {
			hello = {
				result = {
					status = "OK"
				}
			}
		}
	}
	local timestamp = os.time()
	local out = {}
	function write_cb(parsed_data)
		-- nonsense, small name for easyness-to-refer-to in assertions
		-- later on
		out.a = parsed_data
		return true
	end
	assert_false(
		save_report_data(
			old_service_data,
			result,
			object,
			res,
			timestamp,
			write_cb
		)
	)
end

function teardown()
	-- Teardown after each test
	nonstdout = nil
end
